import os
import shutil
import tempfile
from contextlib import chdir
from pathlib import Path

import pytest
from configuraptor.singleton import Singleton

from src.edwh_migrate import migrate


@pytest.fixture
def clean_migrate():
    migrate.registered_functions = {}
    Singleton.clear()  # clean cached Config


@pytest.fixture(scope="session")
def tmp_sqlite_folder(tmp_path_factory):
    return tmp_path_factory.mktemp("sqlite3.tmp")


@pytest.fixture(scope="session")
def sqlite_empty():
    return Path(__file__).parent


@pytest.fixture
def tmp_sqlite_sql_file(tmp_sqlite_folder, sqlite_empty):
    dst = tmp_sqlite_folder / "test.sql"
    shutil.copy(sqlite_empty / "test.sql", dst)
    yield dst
    dst.unlink(missing_ok=True)


@pytest.fixture
def tmp_empty_sqlite_db_file(tmp_sqlite_folder, sqlite_empty):
    dst = tmp_sqlite_folder / "empty_sqlite.db"
    shutil.copy(sqlite_empty / "sqlite_empty" / "empty_sqlite.db", dst)
    os.environ["MIGRATE_URI"] = f"sqlite://{str(dst)}"
    yield dst
    dst.unlink(missing_ok=True)
    del os.environ["MIGRATE_URI"]


@pytest.fixture
def tmp_just_implemented_features_sqlite_sql_file(tmp_sqlite_folder, sqlite_empty):
    return sqlite_empty / "sqlite_empty" / "just_implemented_features.sql"


@pytest.fixture
def tmp_just_implemented_features_sqlite_db_file(tmp_sqlite_folder, sqlite_empty):
    dst = tmp_sqlite_folder / "just_implemented_features.db"
    shutil.copy(sqlite_empty / "sqlite_empty" / "just_implemented_features.db", dst)
    os.environ["MIGRATE_URI"] = f"sqlite://{str(dst)}"
    yield dst
    dst.unlink(missing_ok=True)
    del os.environ["MIGRATE_URI"]


@pytest.fixture()
def fixture_temp_chdir():
    with tempfile.TemporaryDirectory() as _dir:
        cwd = Path(_dir)
        with chdir(cwd):
            yield cwd



from pydal import DAL

from edwh_migrate import migration


@migration
def create_users_table_20231109_001(db: DAL):
    db.executesql(
        """
        CREATE TABLE users (
            id SERIAL PRIMARY KEY,
            username VARCHAR(50) NOT NULL,
            email VARCHAR(100) NOT NULL,
            created_at TIMESTAMP DEFAULT current_timestamp
        )
    """
    )
    db.commit()
    return True


@migration(requires=create_users_table_20231109_001)
def add_password_column_to_users_20231109_001(db: DAL):
    db.executesql(
        """
        ALTER TABLE users
        ADD COLUMN password VARCHAR(100) NOT NULL
    """
    )
    db.commit()
    return True



# coding=utf-8
# SPDX-FileCopyrightText: 2023-present Remco <remco@educationwarehouse.nl>
#
# SPDX-License-Identifier: MIT
# type: ignore

from edwh_migrate import migration


@migration
def feature_1(db):
    # mogelijkheden om in een feature te doen:
    # 1. SQL strings uitvoeren richting database
    # 2. Bestanden lezen en uitvoeren als sql string, of verschillende statements achter elkaar
    # 3. PG_RESTORE uitvoeren mbv plumbum om zo een hele database te recoveren
    # en daarnaast
    # 1. wijzigingen die in de data horen nav een schemawijziging...
    #
    # een andere manier zou zijn om dit uit bestanden te lezen en die door te voeren
    # dan heb je de functies niet zo zeer nodig, maar heb je wel je schema bijgewerkt. Zolang je geen
    # logica van python nodig hebt kun je best veel doen met alleen maar SQL scripts
    # sorted(/features/*.sql) | cat vanaf invoer |
    return True


@migration(requires=feature_1)
def feature_2(db):
    print("feature_2")
    return True


@migration(requires=feature_2)
def feature_3(db):
    print("feature_3")
    return True


@migration
def functionalname_date_sequencenr(db):
    db.executesql("""
    
    """)
    db.commit()
    return True



# coding=utf-8
# SPDX-FileCopyrightText: 2023-present Remco <remco@educationwarehouse.nl>
#
# SPDX-License-Identifier: MIT
"""
This file contains most of the (core) code.

When writing new tasks, make sure:
 * ! executesql() returns tuples, explicitly use as_dict or as_ordered_dict to have easy acces for results
 * ! arguments and placeholders use the `where gid = %(gid)s` syntax, where gid should be a key in the `placeholders`
   argument given to executesql
 * ! make sure to use schema reference:
   - public."user"
   - public.item
   forget and fail...
"""

import contextlib
import datetime
import importlib
import os
import pathlib
import sqlite3
import sys
import time
import typing
import urllib
import urllib.parse
from collections import OrderedDict
from functools import wraps
from pathlib import Path
from typing import Optional

import configuraptor
import dotenv
import plumbum
import psycopg2
import psycopg2.errors
import redis
from configuraptor import alias, asdict, postpone
from configuraptor.errors import ConfigErrorMissingKey, IsPostponedError
from dotenv import find_dotenv
from pydal import DAL, Field
from pydal.objects import Table

try:
    from typedal import TypeDAL
except ImportError:
    TypeDAL = DAL  # type: ignore

registered_functions = OrderedDict()


class BaseEDWHMigrateException(BaseException):
    """
    Most top-level exception class for this module.

    Not caught by `except Exception`.
    """


class EDWHMigrateException(BaseEDWHMigrateException, Exception):
    """
    Common exception class for this module.

    Is caught by `except Exception`.
    """


class UnknownConfigException(BaseEDWHMigrateException):
    """
    Not currently used.
    """


class InvalidConfigException(BaseEDWHMigrateException):
    """
    Thrown when not all requires settings (e.g. database uri) can be found.
    """


class DatabaseNotYetInitialized(BaseEDWHMigrateException):
    """
    Thrown when trying to access the implemented features table before setting it up.
    """


class RequimentsNotMet(BaseEDWHMigrateException):
    """
    Thrown when a migration depends on some requirements, which have not been succesfully installed yet.
    """


class MigrateLockExists(EDWHMigrateException):
    """
    Thrown when trying to migrate but a lock file is already present.

    This indicates the migration has already happened.
    """


class MigrationFailed(EDWHMigrateException):
    """
    Raised when a migration fails for any reason.
    """


class Config(configuraptor.TypedConfig, configuraptor.Singleton):
    """
    These options can be set via pyproject.toml or .env (or a combination).

    Use the [tool.migrate] key.
    Either - or _ can be used in the keys and they can be in any case.
    """

    migrate_uri: str = postpone()
    schema_version: Optional[str]
    redis_host: Optional[str]
    migrate_cat_command: Optional[str]
    database_to_restore: str = "/data/database_to_restore.sql"
    migrate_table: str = "ewh_implemented_features"
    flag_location: str = "/flags"
    schema: str | bool = "public"
    create_flag_location: bool = False
    db_folder: Optional[str] = None
    migrations_file: str = "migrations.py"
    use_typedal: bool = False

    db_uri: str = alias("migrate_uri")

    def __repr__(self) -> str:
        """
        Represent the class by dumping its data.
        """
        data = asdict(self, with_top_level_key=False)
        return f"<Config{data}>"


def _get_config() -> Config:
    """
    First try config from env, then fallback to pyproject.
    """
    dotenv.load_dotenv(find_dotenv(usecwd=True))
    try:
        return Config.load("pyproject.toml", key="tool.migrate")
    except (configuraptor.errors.ConfigError, FileNotFoundError):
        return Config.from_env(load_dotenv=False)


def get_config() -> Config:
    """
    Get the whole config or a specific key.
    """
    config = _get_config()
    config.update_from_env()  # extend toml defaults with env secrets

    return config


T_Dal = typing.TypeVar("T_Dal", bound=DAL)


def define_ewh_implemented_features(db: DAL, rname: str = "ewh_implemented_features") -> Table:
    """
    Define the required table.

    Can be used by pydal2sql to generate a CREATE TABLE statement:
    Example:
        pydal2sql create src/edwh_migrate/migrate.py
            --magic
            --function 'define_ewh_implemented_features(db, "custom_rname")'
            --dialect sqlite
            --format edwh-migrate
    """
    return db.define_table(
        "ewh_implemented_features",
        Field("name", unique=True),
        Field("installed", "boolean", default=False),
        Field("last_update_dttm", "datetime", default=datetime.datetime.now),
        rname=rname,
    )


@typing.overload
def setup_db(
    migrate: bool = False,
    migrate_enabled: bool = False,
    appname: str = " ".join(sys.argv),
    long_running: bool | int = False,
    dal_class: None = None,
    impl_feat_table_name: Optional[str] = None,
    folder: Optional[str] = None,
    config: Optional[Config] = None,
    remove_migrate_tablefile: bool = False,
) -> DAL:
    """
    If `dal_class` is not filled in, a normal DAL instance is returned.
    """


@typing.overload
def setup_db(
    migrate: bool = False,
    migrate_enabled: bool = False,
    appname: str = " ".join(sys.argv),
    long_running: bool | int = False,
    dal_class: typing.Type[T_Dal] = DAL,
    impl_feat_table_name: Optional[str] = None,
    folder: Optional[str] = None,
    config: Optional[Config] = None,
    remove_migrate_tablefile: bool = False,
) -> T_Dal:
    """
    If `dal_class` is passed, an instance of that class will be returned.
    """


def setup_db(
    migrate: bool = False,
    migrate_enabled: bool = False,
    appname: str = " ".join(sys.argv),
    long_running: bool | int = False,
    dal_class: typing.Type[T_Dal] | None = None,
    impl_feat_table_name: Optional[str] = None,
    folder: Optional[str] = None,
    config: Optional[Config] = None,
    remove_migrate_tablefile: bool = False,
) -> T_Dal | DAL:
    """
    Connect to the database and return a DAL object.

    When using postgres, the application name is set to the appname argument, long lasting connections are enabled
    when using long_running, to avoid PGPOOL closing the connection during unpacking of a larger backup file.

    This function will also search for the ewh_implemented_features table,
    which is used to keep track of which migrations have been attempted and applied.

    If this table is not found, an DatabaseNotYetInitialized exception is raised. This is handled in the
    activate_migrations function to allow for the database to be restored from a backup before continuing.

    While migrating, functions registered with the @migration(requires=...) decorator applied, will reconnect using
    `setup_db()` without any arguments.

    :param migrate: migrate is normally turned off, but if you want to run migrations, set this to True
    :param migrate_enabled: normally migrate is turned off, but if you want to run migrations, set this to True
    :param appname: name of this application to register on postgres connections, default: " ".join(sys.argv)
    :param long_running: bool or int to indicate the number of seconds to keep the connection alive
        using pgpool set_client_idle_limit
    :param dal_class: optional DAL class, will use DAL if not given
    :param impl_feat_table_name: optional custom table name for ewh_implemented_features
    :param folder: directory to store sql log, table files etc
    :param config: existing Config object. If not passed, get_config will be called.
    :param remove_migrate_tablefile: remove the ewh_implemented_features.table file if it exists?

    :return: database connection
    """
    try:
        config = config or get_config()

        uri = config.migrate_uri
    except (KeyError, ConfigErrorMissingKey, IsPostponedError) as e:
        raise InvalidConfigException("$MIGRATE_URI not found in environment.") from e

    if dal_class is None:
        # default: pydal.DAL
        # (alternatives: py4web.core.DAL, typedal.TypeDAL, typedal.py4web.DAL)
        dal_class = TypeDAL if config.use_typedal else DAL

    is_postgres = uri.startswith("postgres")
    driver_args: dict[str, typing.Any] = {}
    if is_postgres:
        driver_args["application_name"] = appname
        if not long_running:
            driver_args["keepalives"] = 1

    db = dal_class(
        uri,
        migrate=migrate,
        migrate_enabled=migrate_enabled,
        driver_args=driver_args,
        pool_size=1,
        folder=folder or config.db_folder,
    )

    if is_postgres and long_running:
        # https://www.pgpool.net/docs/latest/en/html/sql-pgpool-set.html
        # make this connection able to live longer, because the functions can take over 30s
        # to perform the job.
        # db.executesql("PGPOOL SET client_idle_limit = 3600;")
        with contextlib.suppress(Exception):
            print("Setting up for long running connection")
            db.executesql(f"PGPOOL SET client_idle_limit = {long_running if str(long_running).isdigit() else 3600};")
            db.rollback()

    if remove_migrate_tablefile:
        tablefile_path = Path(db._folder) / f"{db._uri_hash}_ewh_implemented_features.table"
        tablefile_path.unlink(missing_ok=True)

    define_ewh_implemented_features(db, impl_feat_table_name or config.migrate_table)
    db.commit()

    try:
        db(db.ewh_implemented_features).count()
    except (psycopg2.errors.UndefinedTable, sqlite3.OperationalError) as e:
        db.rollback()
        raise DatabaseNotYetInitialized(f"{config.migrate_table} is missing.", db) from e

    return db


Migration: typing.TypeAlias = typing.Callable[[DAL], bool]


@typing.overload
def migration(
    func: None = None,
    requires: list[Migration] | Migration | None = None,
) -> typing.Callable[[Migration], Migration]:
    """
    Allows calling the decorator with parentheses.

    Example:
        @migration()
        def my_migration_1(): ...
    """


@typing.overload
def migration(
    func: Migration,
    requires: list[Migration] | Migration | None = None,
) -> Migration:
    """
    Allows calling @migration without parens.
    """


def migration(
    func: Migration | None = None,
    requires: list[Migration] | Migration | None = None,
) -> Migration | typing.Callable[[Migration], Migration]:
    """
    Decorator to register a function as a migration function.

    :param func: function to register as a migration function
    :param requires: list of function names that need to be applied before this function can be applied.

    Example:
        @migration
        def my_migration_function(db):
            db.executesql("select 1")
            return True # or False, if the migration failed. On true db.commit() will be performed.
    """
    if func is None and requires:
        # requires is given, so return the decorator that will test if the requirements are met before
        # executing the decorated function, when it's time to really execute the function.

        if callable(requires):  # noqa: SIM108
            # when a single requirement is given, and it is a function, take the name of the function
            required_names = [requires.__name__]
        else:
            # if requires is not callable, then it must be a list of functions, so take the names of those functions.
            required_names = [_.__name__ for _ in requires]

        def decorator(decorated: Migration) -> Migration:
            @wraps(decorated)
            def with_requires(*p: typing.Any, **kwp: typing.Any) -> bool:
                # check requirements
                db = setup_db()
                installed = (
                    db(
                        db.ewh_implemented_features.name.belongs(required_names)
                        & (db.ewh_implemented_features.installed == True)
                    )
                    .select("installed")
                    .column("installed")
                )

                # check if all requirements are in the list of installed features
                if len(installed) != len(required_names) or not all(installed):
                    db.close()
                    print(decorated.__name__, "REQUIREMENTS NOT MET")
                    raise RequimentsNotMet("requirements not met")

                return decorated(*p, **kwp)

            registered_functions[decorated.__name__] = with_requires
            return with_requires

        return decorator

    if func:
        registered_functions[func.__name__] = func
        return func

    return migration


def should_run(db: DAL, name: str) -> bool:
    """
    Checks if a migration function should be run.

    :param db: database connection
    :param name: name of the migration function
    :return: whether the migration function should be run (if not installed) or not (when installed).
    """
    resultset = db(db.ewh_implemented_features.name == name).select()
    row = resultset.first() if resultset else None
    return row.installed is False if row else True


def recover_database_from_backup(set_schema: Optional[str | bool] = None, config: Optional[Config] = None) -> None:
    """
    Recover a database backup.

    Handles 3 situations:
    a) /data/database_to_restore.sql exists:
       just recover the database (uses cat)
    b) /data/database_to_restore.xz exists:
       unpacks and recovers the file (uses xzcat)
    c) /data/database_to_restore.gz exists:
       unpacks and recovers the file (uses zcat)
    d) no file exists:
        tries to recover using an un7zipped file,
        possibly downloaded from backblaze, but this code
        is stale.

    Effectively running something like:
        A,B)  7z x -so -p"secret" db.7z | psql -h  127.0.0.1 -U postgres -d backend
          C) cat /data/database_to_restore.sql | psql -h  127.0.0.1 -U postgres -d backend

    """
    config = config or get_config()
    set_schema = set_schema or config.schema

    print("RECOVER_DATABASE_FROM_BACKEND started ")
    prepared_sql_path = pathlib.Path(config.database_to_restore)

    if not prepared_sql_path.exists():
        raise FileNotFoundError(prepared_sql_path)
    extension = prepared_sql_path.suffix.lower()

    cat_command = {
        ".sql": "cat",
        ".xz": "xzcat",
        ".gz": "zcat",
    }.get(extension, config.migrate_cat_command)
    if not cat_command:
        raise NotImplementedError(f"Extension {extension} not supported for {prepared_sql_path}")
    unpack = plumbum.local[cat_command][prepared_sql_path]

    # unpack is now the command to cat the sql to stdout
    # which is piped to psql to perform the recovery

    # parse the db uri to find hostname and port and such to
    # feed to pqsl as commandline arguments
    uri = urllib.parse.urlparse(config.migrate_uri)

    if is_postgres := uri.scheme.startswith("postgres"):
        # prepare the psql command
        psql = plumbum.local["psql"][config.migrate_uri]
        sql_consumer = psql
    else:
        filepath = uri.path
        if "///" not in config.migrate_uri:
            # else absolute path, don't strip!
            filepath = filepath.strip("/")
        sqlite_database_path = pathlib.Path(uri.netloc) / pathlib.Path(filepath)
        sql_consumer = plumbum.local["sqlite3"][sqlite_database_path]
    # combine both
    cmd = unpack | sql_consumer

    print("UNPACKING AND INSTALLING DATABASE:", cmd)
    # noinspection PyStatementEffect
    # is plumbum syntax
    cmd() > "/dev/null"
    print("Done unpacking and feeding to", cmd)

    if is_postgres and set_schema:
        echo = plumbum.local["echo"]
        cmd = echo[f"SET search_path TO {set_schema};"] | sql_consumer
        print("For postgres: set schema to public:", cmd)
        cmd()


def try_setup_db() -> typing.Optional[DAL]:
    """
    Handle multiple scenario's such as an existing db, a db that can be loaded from a backup or a fully new db.
    """
    started = time.time()
    while time.time() - started < 600:
        try:
            db = setup_db()
            print("activate_migrations connected after", time.time() - started, "seconds.")

            db.commit()
            # without an error, all *should* be well...
            return db
        except DatabaseNotYetInitialized as e:
            # the connection succeeded, but when encountering an empty databse,
            # the features table will not be found, and this will raise an exception
            # this error is to be expected, since the database is not yet recovered...
            print("table not found, starting database restore")
            # recover the erroneous database connection to start database recovery.
            _, db = e.args
            # clear the current connection from errors :
            db.rollback()
            try:
                print("RECOVER: attempting recovery from a backup")
                db.close()
                recover_database_from_backup()
                # give db-0 a while to catch up
                time.sleep(10)
                # reconnect to the database, since the above operation takes longer than timeout period on dev machines.
                return setup_db()

            except FileNotFoundError as e:
                with contextlib.suppress(DatabaseNotYetInitialized):
                    print(f"RECOVER: {e} not found. Starting from scratch.")
                    return setup_db(migrate=True, migrate_enabled=True)
                with contextlib.suppress(DatabaseNotYetInitialized):
                    print("RECOVER: Failed. Starting from scratch with new .table file.")
                    return setup_db(migrate=True, migrate_enabled=True, remove_migrate_tablefile=True)

            except Exception as e:
                print("RECOVER: database recovery went wrong:", e)
            break  # don't retry
        except Exception as e:
            print(
                "activate_migrations failed to connect to database. "
                "sleeping and retrying in 3 seconds. will try for 10 min."
            )
            print(e)
            time.sleep(3)

    # shouldn't happen :(
    return None


def activate_migrations(config: Optional[Config] = None) -> bool:
    """
    Start the migration process, don't wait for a lock.
    """
    db = try_setup_db()
    if not db:
        raise ValueError("No db could be set up!")

    config = config or get_config()

    successes = []
    # perform migrations
    for name, function in registered_functions.items():
        print("test:", name)
        if should_run(db, name):
            print("run: ", name)
            # create a new database connection
            # because there could be tables being defined,
            # and we want all the functions to be siloed and not
            # have database schema dependencies and collisions.
            db_for_this_function = setup_db()
            result = function(db_for_this_function)
            successes.append(result)
            if result:
                # commit the change to db
                db_for_this_function.commit()
                # close this specific cursor, it's not used any longer.
                db_for_this_function.close()
                print("ran: ", name, "successfully. ")
                # alleen bij success opslaan dat er de feature beschikbaar is
                db.ewh_implemented_features.update_or_insert(
                    db.ewh_implemented_features.name == name,
                    name=name,
                    installed=True,
                    last_update_dttm=datetime.datetime.now(),
                )
            else:
                print("ran: ", name, " and failed. ")
                successes.append(False)
                # try a rollback, because we should ignore whatever happend
                db_for_this_function.rollback()
                # and close because this connection is not used any longer.
                db_for_this_function.close()
                db.ewh_implemented_features.update_or_insert(
                    db.ewh_implemented_features.name == name,
                    name=name,
                    installed=False,
                    last_update_dttm=datetime.datetime.now(),
                )
            db.commit()
        else:
            print("already installed. ")

    db.close()

    # clean redis whenever possible
    # reads REDIS_MASTER_HOST from the environment
    if redis_host := config.redis_host:
        r = redis.Redis(redis_host)
        keys = r.keys()
        print(f"Removing {len(keys)} keys from redis.")
        for key in keys:
            del r[key]
        print("done")
    return all(successes)


@contextlib.contextmanager
def schema_versioned_lock_file(
    flag_location: str | Path | None = None, create_flag_location: bool = False, config: Optional[Config] = None
) -> typing.Generator[Path | None, None, None]:
    """
    Context manager that creates a lock file for the current schema version.
    """
    config = config or get_config()

    flag_location = Path(flag_location or config.flag_location)
    if not flag_location.exists():
        if create_flag_location or config.create_flag_location:
            flag_location.mkdir()
        else:
            raise NotADirectoryError(
                f"Flag directory {flag_location} does not exist. "
                f"Please create it or set `create_flag_location` to True."
            )

    if (schema_version := config.schema_version) is None:
        print("No schema version found, ignoring any lock files.")
        yield None
    else:
        print("testing migrate lock file with the current version")
        lock_file = flag_location / f"migrate-{schema_version}.complete"
        print("Using lock file: ", lock_file)
        if lock_file.exists():
            print("migrate: lock file already exists, migration should be completed. Aborting migration")
            raise MigrateLockExists(str(lock_file))
        else:
            # create the lock asap, to avoid racing conditions with other possible migration processes
            lock_file.touch()
            try:
                yield lock_file
            except MigrationFailed:
                # remove the lock file, so that the migration can be retried.
                print("ERROR: migration failed, removing the lock file.")
                print(f"Check the {config.migrate_table} table for details.")
                lock_file.unlink()

            except BaseException:
                # since another exception was raised, reraise it for the stack trace.
                lock_file.unlink()
                raise


def _console_hook(args: list[str], config: Optional[Config] = None) -> None:  # pragma: no cover
    if "-h" in args or "--help" in args:
        print(
            """
        Execute migrate to run the migration in `migrations.py` from the current working directory.

        The database connection url is read from the MIGRATE_URI environment variable.
        It should be a pydal compatible connection string.


        ## Testing migrations using sqlite

        Create a test setup using sqlite3:
        $sqlite3 test.db  "create table ewh_implemented_features(id, name, installed, last_update_dttm);"
        $export MIGRATE_URI='sqlite://test.db'

        """
        )
        exit(0)

    config = config or get_config()

    # get the versioned lock file path, as the config performs the environment variable expansion
    with contextlib.suppress(MigrateLockExists), schema_versioned_lock_file(config=config):
        if args:
            print(f"Using argument {args[0]} as a reference to the migrations file.")
            # use the first argument as a reference to the migrations file
            # or the folder where the migrations file is stored
            arg = Path(args[0])
            if arg.exists() and arg.is_file():
                print(f"importing migrations from {arg}")
                sys.path.insert(0, str(arg.parent))
                # importing the migrations.py file will register the functions
                importlib.import_module(arg.stem)
            elif arg.exists() and arg.is_dir():
                print(f"importing migrations from {arg}/migrations.py")
                sys.path.insert(0, str(arg))
                # importing the migrations.py file will register the functions
                importlib.import_module("migrations")
            else:
                print(f"ERROR: no migrations found at {arg}", file=sys.stderr)
                exit(1)
        elif config.migrations_file and (arg := Path(config.migrations_file)) and arg.exists():
            print(f"importing migrations from {arg}")
            sys.path.insert(0, str(arg.parent))
            # importing the migrations.py file will register the functions
            importlib.import_module(arg.stem)

        elif Path("migrations.py").exists():
            print("migrations.py exists, importing @migration decorated functions.")
            sys.path.insert(0, os.getcwd())
            # importing the migrations.py file will register the functions
            import migrations  # noqa F401: semantic import here
        else:
            print(f"ERROR: no migrations found at {os.getcwd()}", file=sys.stderr)
            exit(1)
        print("starting migrate hook")
        print(f"{len(registered_functions)} migrations discovered")
        if activate_migrations():
            print("migration completed successfully, marking success.")
        else:
            raise MigrationFailed("Not every migration succeeded.")


def console_hook() -> None:  # pragma: no cover
    """
    Activated by migrate shell script, sets a lock file before activate_migrations.

    lockfile: '/flags/migrate-{os.environ["SCHEMA_VERSION"]}.complete'
    """
    _console_hook(sys.argv[1:])


# ------------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------------
# ----------------------------- registered functions ---------------------------------------------
# ------------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------------

# @migration
# def start_from_scratch(db: DAL) -> bool:
#     print(f"String from scratch using {db}")
#     return True



import os
import pathlib

import plumbum
import pydal
import pytest
from contextlib_chdir import chdir
from pydal import DAL

from src.edwh_migrate import migrate, recover_database_from_backup
from src.edwh_migrate.__about__ import __version__
from src.edwh_migrate.migrate import (
    MigrateLockExists,
    MigrationFailed,
    get_config,
    schema_versioned_lock_file,
)
from .fixtures import (  # noqa
    clean_migrate,
    fixture_temp_chdir,
    sqlite_empty,
    tmp_empty_sqlite_db_file,
    tmp_just_implemented_features_sqlite_db_file,
    tmp_just_implemented_features_sqlite_sql_file,
    tmp_sqlite_folder,
    tmp_sqlite_sql_file,
)


def test_version():
    assert isinstance(__version__, str)
    assert __version__


def test_env_migrate_uri_is_missing():
    with chdir("tests"):
        with pytest.raises(migrate.InvalidConfigException):
            migrate.setup_db()


def test_migrate_from_toml(clean_migrate):
    config = get_config()
    assert migrate.setup_db(migrate=True, migrate_enabled=True)

    assert config.migrate_uri == "sqlite:///tmp/migrate_example.sqlite"

    assert pathlib.Path("/tmp/migrate_example.sqlite").exists()


def test_setup_db_custom_pydal_class(clean_migrate):
    class DALish(DAL):
        ...

    db = migrate.setup_db(dal_class=DALish)

    assert isinstance(db, DALish)


def test_apply_empty_run_to_empty_sqlite(tmp_empty_sqlite_db_file, clean_migrate):
    with pytest.raises(migrate.DatabaseNotYetInitialized):
        migrate.setup_db()


def test_starts_without_registered_migrations():
    assert len(migrate.registered_functions) == 0, "No migrations should be registered"


def test_registration_works():
    @migrate.migration
    def dummy(db):
        return True

    assert len(migrate.registered_functions) == 1, "only one function should be registered"


def dump_db(db: pydal.DAL, *, echo=False):
    output = plumbum.local["sqlite3"][db._uri.split("://")[1]][".dump"]()
    if echo:
        print(output)
    return output


def test_always_true_dummy_is_migrated(tmp_just_implemented_features_sqlite_db_file):
    @migrate.migration()
    def dummy(db):
        return True

    assert len(migrate.registered_functions) == 1, "exactly one function should be registerend"
    result = migrate.activate_migrations()
    assert result is True, "the dummy returning True should have been marked as successful"
    db = migrate.setup_db()
    # dump_db(db, echo=True)
    assert db(db.ewh_implemented_features).count() == 1, "exactly one row should be in the table"
    rs = db(db.ewh_implemented_features).select()
    assert len(rs) == 1, "exactly one row should be in the table"
    assert rs.first().name == "dummy", "the name of the row should be dummy"
    assert rs.first().installed is True, "the row should be marked as installed in the database"


def test_dummy_is_not_migrated_twice(tmp_just_implemented_features_sqlite_db_file, capsys):
    @migrate.migration
    def dummy(db):
        return True

    assert len(migrate.registered_functions) == 1, "exactly one function should be registerend"
    result = migrate.activate_migrations()
    assert result is True, "the dummy returning True should have been marked as successful"
    result = migrate.activate_migrations()
    assert (
        "already installed." in capsys.readouterr().out
    ), "the dummy returning True should have been marked as successful"
    db = migrate.setup_db()
    # dump_db(db, echo=True)
    assert db(db.ewh_implemented_features).count() == 1, "exactly one row should be in the table"
    rs = db(db.ewh_implemented_features).select()
    assert len(rs) == 1, "exactly one row should be in the table"
    assert rs.first().name == "dummy", "the name of the row should be dummy"
    assert rs.first().installed is True, "the row should be marked as installed in the database"


@pytest.mark.parametrize("scenario", ["as_list", "as_function"])
def test_dependencies(clean_migrate, tmp_just_implemented_features_sqlite_db_file, capsys, scenario):
    @migrate.migration
    def required(db):
        return True

    print(migrate.registered_functions)
    assert len(migrate.registered_functions) == 1
    result = migrate.activate_migrations()
    assert result is True, "the required migration returning True should have been marked as successful"

    @migrate.migration(requires=required if scenario == "as_function" else [required])
    def dependent(db):
        return True

    assert len(migrate.registered_functions) == 2
    result = migrate.activate_migrations()
    assert result is True

    db = migrate.setup_db()
    assert result is True, "the dependent returning True should have been marked as successful"

    assert db(db.ewh_implemented_features.installed == True).count() == 2, "exactly two rows should be marked installed"


@pytest.mark.parametrize("scenario", ["as_list", "as_function"])
def test_dependency_failure(clean_migrate, tmp_just_implemented_features_sqlite_db_file, capsys, scenario):
    @migrate.migration
    def required(db):
        return False

    @migrate.migration(requires=required if scenario == "as_function" else [required])
    def dependent(db):
        return True

    assert len(migrate.registered_functions) == 2

    with pytest.raises(migrate.RequimentsNotMet):
        migrate.activate_migrations()

    db = migrate.setup_db()
    dump_db(db, echo=True)
    assert db(db.ewh_implemented_features.installed == True).count() == 0, "requirement failed, no succes possible"
    assert (
        db(db.ewh_implemented_features.installed == False).count() == 1
    ), "because of the exception, `dependent` is never written to the database. "


def test_recover_database_from_backup(tmp_just_implemented_features_sqlite_sql_file, tmp_empty_sqlite_db_file):
    config = get_config()

    with pytest.raises(migrate.DatabaseNotYetInitialized):
        migrate.setup_db(migrate=False)

    fake_path = pathlib.Path("/tmp/fake_file.xyz")
    fake_path.unlink(missing_ok=True)
    config.database_to_restore = str(fake_path)

    with pytest.raises(FileNotFoundError):
        recover_database_from_backup()

    fake_path.touch()
    with pytest.raises(NotImplementedError):
        # invalid extension
        recover_database_from_backup()

    config.database_to_restore = str(tmp_just_implemented_features_sqlite_sql_file)
    recover_database_from_backup()

    # no error anymore:
    assert migrate.setup_db(migrate=False)


def test_config():
    config = get_config()
    assert config is get_config()  # singleton

    assert config.database_to_restore is get_config().database_to_restore

    assert "<Config{" in repr(config)


def test_schema_versioned_lock_file(capsys, clean_migrate):
    config = get_config()

    config.flag_location = "/tmp/test_flag_dir"

    flag_dir = pathlib.Path(config.flag_location)
    if flag_dir.exists():
        [_.unlink() for _ in flag_dir.glob("*")]
        flag_dir.rmdir()

    with pytest.raises(NotADirectoryError):
        with schema_versioned_lock_file() as lock:
            pass

    config.schema_version = None

    with schema_versioned_lock_file(flag_location=flag_dir, create_flag_location=True) as lock:
        assert lock is None

    config.schema_version = "1"

    with schema_versioned_lock_file(flag_location=flag_dir, create_flag_location=True) as lock:
        assert lock

    with pytest.raises(MigrateLockExists):
        with schema_versioned_lock_file(flag_location=flag_dir, create_flag_location=False) as lock:
            pass

    # config.schema_version = "2"
    os.environ["SCHEMA-VERSION"] = "2"
    config.update_from_env()

    assert config.schema_version == "2"

    with schema_versioned_lock_file(flag_location=flag_dir, create_flag_location=False) as lock:
        raise MigrationFailed()

    captured = capsys.readouterr()

    assert "removing the lock file" in captured.out

    with pytest.raises(Exception):
        with schema_versioned_lock_file(flag_location=flag_dir, create_flag_location=False) as lock:
            raise Exception()

    captured = capsys.readouterr()

    assert "removing the lock file" not in captured.out


def test_without_migrate_uri_but_with_db_uri_and_folder(fixture_temp_chdir, clean_migrate):
    os.environ["DB_URI"] = "sqlite://storage.sqlite"

    db_folder = fixture_temp_chdir / "database"
    db_folder.mkdir()

    os.environ["DB_FOLDER"] = str(db_folder)

    try:
        migrate.setup_db(migrate_enabled=True, migrate=True)

        assert db_folder.exists()

        assert plumbum.local["ls"](os.getcwd()).strip() == "database"

        db_ls = plumbum.local["ls"](db_folder)

        assert "ewh_implemented_features.table" in db_ls
        assert "sql.log" in db_ls
        assert "storage.sqlite" in db_ls
    finally:
        del os.environ["DB_URI"]
        del os.environ["DB_FOLDER"]
